# -*- coding: utf-8 -*-
from robogame_engine import GameObject
from robogame_engine.constants import ROTATE_TURNING
from robogame_engine.geometry import Vector
from robogame_engine.theme import theme


class Projectile(GameObject):
    """описывает поведение полета снаряда."""
    coord = None
    rotate_mode = ROTATE_TURNING
    radius = 1
    selectable = False
    layer = 3

    death_animation = None
    max_distance = 100.0

    def __init__(self, owner=None, speed=None, ttl=None, attached_ttl=None, **kwargs):
        super(Projectile, self).__init__(**kwargs)
        self.__initial_coord = owner.coord.copy()
        self._owner = owner
        self.__ttl = ttl
        self.__speed = speed
        self.__attached = None
        self.__attached_ttl = attached_ttl

    @property
    def sprite_filename(self):
        return "not-found.png"

    @property
    def owner(self):
        return self._owner

    @property
    def ttl(self):
        return self.__ttl

    @property
    def has_hit(self):
        return self.__attached is not None

    @property
    def attached(self):
        return self.__attached

    @property
    def zoom(self):
        if self.has_hit:
            return self.__attached.zoom
        return 0.5

    @property
    def is_alive(self):
        return self.__ttl > 0

    def game_step(self):
        if not theme.DRONES_CAN_FIGHT:
            return

        if self.has_hit:
            self.__attached.game_step()
            return

        if self.is_alive:
            self.__ttl = max(self.__ttl - 1, 0)
            # Обычный цикл полета пока живы
            super(Projectile, self).game_step()
        else:
            # Останавливаемся если отжили цикл полета
            # Используем прямой вызов останова, вместо события
            self.state.stop()
            # Промах, никакой анимации
            self.scene.remove_object(self)

    def turn_to(self, target, speed=None):
        # Снаряд не управляемый
        pass

    def move_at(self, target, speed=None):
        # Двигаемся только по прямой, незачем иметь возможность изменять маршрут
        pass

    def on_stop(self):
        pass

    def on_born(self):
        if not theme.DRONES_CAN_FIGHT:
            self.scene.remove_object(self)
            return
        vector = Vector.from_direction(self._owner.direction, module=PlasmaProjectile.max_distance)
        point = self._owner.coord.copy() + vector
        super(Projectile, self).move_at(point, speed=self.__speed)

    def on_overlap_with(self, obj_status):
        if not hasattr(obj_status, "damage_taken"):
            return
        # Пролетаем некомандные объекты
        if obj_status.team is None:
            return
        if theme.TEAM_DRONES_FRIENDLY_FIRE:
            # Не наносим урон себе
            if obj_status.id == self._owner.id:
                return
        else:
            # Пролетаем свои объекты
            if obj_status.team == self._owner.team:
                return
        # За премя жизни ни в кого не попали
        if not obj_status.is_alive or not self.is_alive:
            return
        self.__ttl = 0
        self.stop()
        self.state.stop()
        obj_status.damage_taken(theme.PROJECTILE_DAMAGE)
        if self.death_animation is not None:
            self.__attached = self.death_animation(
                projectile=self, target=obj_status,
                distance=int(obj_status.distance_to(self) / 2),
                direction=Vector.from_points(obj_status.coord, self.coord).direction,
                ttl=self.__attached_ttl
            )


class Gun(object):
    projectile = None

    def __init__(self, owner=None):
        self._owner = owner
        self._cooldown = 0

    @property
    def owner(self):
        return self._owner

    @property
    def can_shot(self):
        return self._cooldown <= 0

    @property
    def shot_distance(self):
        return self.projectile.max_distance

    def shot(self, target):
        if not self._owner.is_alive:
            return
        if not theme.DRONES_CAN_FIGHT or not self.can_shot:
            return
        self._cooldown = theme.PLASMAGUN_COOLDOWN_TIME
        coord = self.owner.coord.copy()
        prtl = self.projectile(coord=coord.copy(), owner=self.owner, direction=self.owner.direction)
        prtl.set_team(self.owner.team)

    def game_step(self):
        # Восстановление после выстрела
        if self._cooldown > 0:
            self._cooldown = max(self._cooldown - theme.PLASMAGUN_COOLDOWN_RATE, 0)


class PlasmaProjectile(Projectile):
    """Используемый снаряд"""
    radius = 15
    max_distance = 580.0

    class __DeathAnimation(object):
        """
            хранит информацию о цели и имееют свое время жизни для
            для поддержания жизнеспособности projectile с анимацией
        """
        def __init__(self, projectile=None, target=None, distance=None, direction=None, ttl=0):
            self.__projectile = projectile
            self.__target = target
            self.__distance = distance
            self.__direction = direction
            self.__initial_target_direction = target.direction
            self.__initial_ttl = ttl
            self.__ttl = ttl

        @property
        def zoom(self):
            t = self.__initial_ttl
            return 1.0 * self.__ttl / t if self.__ttl > 0 else 0.5

        @property
        def sprite_filename(self):
            return "teams/any_projectile_explosion.png"

        @property
        def target(self):
            return self.__target

        @property
        def distance(self):
            return self.__distance

        @property
        def direction(self):
            return self.__direction

        @property
        def ttl(self):
            return self.__ttl

        @property
        def is_alive(self):
            return self.__ttl > 0

        def game_step(self):
            if self.is_alive:
                self.__ttl = max(self.__ttl - self.__projectile._owner.scene.game_speed, 0)
            else:
                return

            # Были прикреплены к оппоненту, анимируем попадание
            direction = - self.__initial_target_direction + self.__target.direction + self.__direction
            vector = Vector.from_direction(direction, module=self.__distance)
            self.__projectile.coord.x = self.__target.x + vector.x
            self.__projectile.coord.y = self.__target.y + vector.y

            if not self.is_alive:
                self.__projectile.scene.remove_object(self.__projectile)

    death_animation = __DeathAnimation

    def __init__(self, **kwargs):
        super(PlasmaProjectile, self).__init__(
            speed=theme.PROJECTILE_SPEED,
            ttl=theme.PROJECTILE_TTL,
            attached_ttl=int(theme.PROJECTILE_TTL / 4), **kwargs
        )

    @property
    def sprite_filename(self):
        if self.is_alive or self.attached is None:
            return "teams/{}/projectile.png".format(self.owner.team)
        return self.attached.sprite_filename

    @property
    def zoom(self):
        if self.has_hit:
            # Эффект попадания в цель
            return self.attached.zoom
        else:
            # эффект появления, чтобы снаряд не возникал из ниоткуда
            showTime = 5
            bornTime = theme.PROJECTILE_TTL - self.ttl
            if bornTime < showTime:
                return 1.0 * (bornTime / showTime)
            # Эффект растворения в космосе в конце дистанции 
            return 1.0 if self.ttl > 10 else 1.0 * self.ttl / 10


# Используемое оружие
class PlasmaGun(Gun):
    projectile = PlasmaProjectile

    def __init__(self, owner=None):
        super(PlasmaGun, self).__init__(owner=owner)



# -*- coding: utf-8 -*-
import random

from robogame_engine.theme import theme

from astrobox.cargo import CargoTransition


class Strategy(object):
    def __init__(self, unit=None, id=None, group=None, is_group_unique=False):
        self.__unit = unit
        self.__id = id
        self.__group = group
        self.__group_unique = is_group_unique

    @property
    def unit(self):
        return self.__unit

    @property
    def id(self):
        return self.__id

    @property
    def group(self):
        return self.__group

    @property
    def is_group_unique(self):
        return self.__group_unique

    def reset(self):
        pass

    @property
    def is_finished(self):
        return False

    def game_step(self):
        pass

    def on_stop(self):
        pass


# Атомарные стратегии
class StrategyCargoLoading(Strategy):
    def __init__(self, cargo_transition, **kwargs):
        kwargs['id'] = "cargo|loading"
        kwargs['group'] = "cargo"
        super(StrategyCargoLoading, self).__init__(kwargs)
        self.__transition = cargo_transition

    @property
    def is_finished(self):
        return self.__transition.is_finished

    def game_step(self):
        self.__transition.game_step()


class StrategyCargoUnloading(Strategy):
    def __init__(self, cargo_transition, **kwargs):
        kwargs['id'] = "cargo|unloading"
        kwargs['group'] = "cargo"
        super(StrategyCargoUnloading, self).__init__(**kwargs)
        self.__transition = cargo_transition

    @property
    def is_finished(self):
        return self.__transition.is_finished

    def game_step(self):
        self.__transition.game_step()


class StrategyApproach(Strategy):
    def __init__(self, target_point=None, distance=0, condition=None, **kwargs):
        if "id" not in kwargs:              kwargs["id"] = "approach"
        if "group" not in kwargs:           kwargs["group"] = "approach"
        if "is_group_unique" not in kwargs: kwargs["is_group_unique"] = True
        super(StrategyApproach, self).__init__(**kwargs)
        self._target_point = target_point
        self._target_distance = distance
        self.__done = False
        self.__last_distance = None
        self.__conditional_approach = condition

    @property
    def is_finished(self):
        # Если усnановлено условие сближения и оно ложно, заканчиваем выполненение
        if self.__conditional_approach is not None and not self.__conditional_approach():
            return True
        return int(self.unit.distance_to(self._target_point)) <= self._target_distance

    def game_step(self):
        new_distance = int(self.unit.distance_to(self._target_point))
        if self.__last_distance is not None and new_distance < self.__last_distance:
            self.__last_distance = new_distance
            return
        if self.unit.is_moving:
            return
        if new_distance > self._target_distance:
            self.__last_distance = new_distance
            self.unit.move_at(self._target_point.copy(), speed=theme.DRONE_SPEED)


# Комбинированные стратегии
class StrategySequence(Strategy):
    def __init__(self, *strategies, **kwargs):
        super(StrategySequence, self).__init__(**kwargs)
        self.__strategies = strategies
        self.__current_strategy = self.__strategies[0]

    def _next_strategy(self):
        if self.__current_strategy is None:
            return False
        strategy_id = self.__strategies.index(self.__current_strategy)
        if strategy_id < 0 or strategy_id + 1 >= len(self.__strategies):
            self.__current_strategy = None
            return False
        self.__current_strategy = self.__strategies[strategy_id + 1]
        return True

    def __str__(self):
        strout = "{} {} {}(".format(self.__class__.__name__, self.unit.__class__.__name__, self.unit)
        for s in self.__strategies:
            strout += ("{}" if self.__strategies.index(s) == 0 else ", {}").format(str(s))
        return strout + ")"

    @property
    def is_finished(self):
        return self.__current_strategy is None

    def game_step(self):
        if self.__current_strategy is None:
            return
        if self.__current_strategy.is_finished:
            if not self._next_strategy():
                return
        self.__current_strategy.game_step()


class StrategyApproachAndLoad(StrategySequence):
    def __init__(self, unit=None, target_unit=None, distance=None, **kwargs):
        if distance is None:
            distance = theme.CARGO_TRANSITION_DISTANCE - 1
        super(StrategyApproachAndLoad, self).__init__(
            StrategyApproach(unit=unit, target_point=target_unit.coord, distance=distance,
                             condition=self.check_target_have_elerium),
            StrategyCargoLoading(CargoTransition(cargo_from=target_unit.cargo, cargo_to=unit.cargo),
                                 unit=unit, is_group_unique=True),
            unit=unit, id="approach&load", group="approach", is_group_unique=True)
        self.__target_unit = target_unit

    def check_target_have_elerium(self):
        return self.__target_unit.cargo.payload > 0


class StrategyApproachAndUnload(StrategySequence):
    def __init__(self, unit=None, target_unit=None, distance=0, **kwargs):
        super(StrategyApproachAndUnload, self).__init__(
            StrategyApproach(unit=unit, target_point=target_unit.coord, distance=distance),
            StrategyCargoUnloading(CargoTransition(cargo_from=unit.cargo, cargo_to=target_unit.cargo),
                                   unit=unit, is_group_unique=True),
            unit=unit, id="approach&unload", group="approach", is_group_unique=True)
        self.__cargo_unloading = None


# Комплексные стратегии
class StrategyHarvesting(Strategy):
    def __init__(self, unit=None):
        super(StrategyHarvesting, self).__init__(unit=unit, id="harvesting", group="harvest", is_group_unique=True)
        self.__substrategy = None
        assert unit is not None
        assert hasattr(unit, 'elerium_stock')

    def anyAsteroid(self):
        return random.choice(self.unit.scene.asteroids)

    def reset(self):
        self.__substrategy = None

    @property
    def current_strategy_id(self):
        if self.__substrategy is not None:
            return self.__substrategy.id
        return ""

    def get_nearest_elerium_stock(self):
        elerium_stocks = [asteriod for asteriod in self.unit.scene.asteroids if asteriod.cargo.payload > 0]
        elerium_stocks += [drone for drone in self.unit.scene.drones if not drone.is_alive and drone.cargo.payload > 0]
        for drone in self.unit.teammates:
            if drone.elerium_stock is not None and \
                    not drone.cargo.is_full and \
                    drone.elerium_stock in elerium_stocks:
                elerium_stocks.remove(drone.elerium_stock)

        if not elerium_stocks:
            return None
        elerium_stocks = sorted(elerium_stocks, key=lambda x: x.distance_to(self.unit))
        return elerium_stocks[0]

    def game_step(self):
        # Даем возможность переопределять выбор источника elerium'а
        nearest_calc = self.unit if hasattr(self.unit, 'get_nearest_elerium_stock') else self
        if self.__substrategy is None or self.__substrategy.is_finished:
            if self.unit.cargo.is_full:
                self.unit.set_elerium_stock(None)
                self.__substrategy = StrategyApproachAndUnload(unit=self.unit, target_unit=self.unit.mothership)
            else:
                near_elerium_stock = nearest_calc.get_nearest_elerium_stock()
                if near_elerium_stock is not None:
                    self.unit.set_elerium_stock(near_elerium_stock)
                    self.__substrategy = StrategyApproachAndLoad(unit=self.unit, target_unit=near_elerium_stock)
                else:
                    if self.unit.cargo.payload > 0 and not self.unit.mothership.cargo.is_full:
                        self.unit.set_elerium_stock(None)
                        self.__substrategy = StrategyApproachAndUnload(unit=self.unit,
                                                                       target_unit=self.unit.mothership)
                    else:
                        # Делаем видимость загруженности дрона работой
                        self.__substrategy = StrategyApproach(unit=self.unit, target_point=self.anyAsteroid().coord)
        if self.__substrategy is not None:
            self.__substrategy.game_step()


class StrategyHunting(Strategy):
    _teams_strategies = {}

    @classmethod
    def getTeamStrategy(cls, team, hunter):
        if team not in cls._teams_strategies:
            cls._teams_strategies[team] = StrategyHunting(unit=hunter, id="hunting", group="hunting",
                                                          is_group_unique=True)
        return cls._teams_strategies[team]

    def __init__(self, **kwargs):
        super(StrategyHunting, self).__init__(**kwargs)
        self._hunters = []
        self._victims = []

    def get_victim(self, hunter):
        if hunter not in self._hunters:
            self._hunters.append(hunter)

        if hunter.victim is not None \
                and hunter.victim.distance_to(hunter.victim.mothership) > theme.MOTHERSHIP_SAFE_DISTANCE \
                and hunter.victim.cargo.payload > 0:
            return hunter.victim

        # Все дроны оппонентов с непустым карго
        enemies = [drone for drone in hunter.scene.drones if
                   drone.team != hunter.team and drone.is_alive and drone.cargo.payload > 0]
        # Дроны оппонетнов дальше, чем дистанция до их mothership-а
        enemies = [enemy for enemy in enemies if enemy.distance_to(enemy.mothership) > theme.MOTHERSHIP_SAFE_DISTANCE]
        for mate in self._hunters:
            if hunter != mate and mate.victim is not None and mate.victim in enemies:
                enemies.remove(mate.victim)
        victim = None
        for enemy in enemies:
            if victim is None or hunter.distance_to(enemy) < hunter.distance_to(victim):
                victim = enemy
        return victim

    def game_step(self, hunter):
        if not hasattr(hunter, 'substrategy') or hunter.substrategy is None:
            hunter.substrategy = StrategyHarvesting(unit=hunter)

        if hunter.victim is not None and (not hunter.victim.is_alive or
                                          int(hunter.victim.distance_to(
                                              hunter.victim.mothership)) < theme.MOTHERSHIP_HEALING_DISTANCE):
            hunter._victim = None
            hunter._victim_stamp = 0

        move_at_point = None
        DRONE_VICTIM_FOCUS_TIME = 4
        if hunter.victim is not None:
            # Счетчик повторного поиска жертвы
            hunter._victim_stamp = min(hunter._victim_stamp + 1, DRONE_VICTIM_FOCUS_TIME)

        # Разгрузимся, чтобы не потерять нажитое
        if hunter.is_unloading:
            # Собираем елериум пока не нашли жертву
            # print(self.__class__.__name__+"::game_step", "bringing elerium to mothership")
            hunter.substrategy.game_step()
            return

        # Получим ближайшую жертву от другого вызова game_step
        if hunter._next_victim is not None:
            if hunter.victim is None:
                move_at_point = hunter.set_victim(hunter._next_victim)

        # Выберем ближашего свободного союзника
        victim = None
        while True:
            victim = self.get_victim(hunter)
            if victim is None:
                break
            victim_distance = victim.distance_to(hunter)
            closertm = [mate for mate in hunter.teammates if
                        mate.victim is None and mate.distance_to(victim) < victim_distance]
            closertm = [mate for mate in closertm if mate._next_victim is None and not mate.is_unloading]
            if not closertm:
                # нет никого свобожного ближе
                break
            closertm = sorted(closertm, key=lambda x: x.distance_to(victim))
            closertm[0]._next_victim = victim
            victim = None
        is_new_victim = victim is not None and hunter.victim != victim

        if is_new_victim:
            move_at_point = hunter.set_victim(victim)
        elif hunter.victim is not None:
            if hunter.state.target_point != hunter.victim.coord and hunter._victim_stamp >= DRONE_VICTIM_FOCUS_TIME:
                # Обновим вектор движения раз в несколько циклов
                hunter._victim_stamp = 0
                move_at_point = hunter.victim.coord.copy()
        if move_at_point is not None and int(hunter.distance_to(move_at_point)) > hunter.radius:
            hunter.move_at(move_at_point.copy(), speed=theme.DRONE_SPEED)

        # Собираем елериум пока не нашли жертву
        if hunter.victim is None and victim is None:
            hunter.substrategy.game_step()


class StrategyDestroyer(Strategy):
    def __init__(self, **kwargs):
        super(StrategyDestroyer, self).__init__(**kwargs)
        self.__substrategy = None
        self._target_unit = None
        self.__done = False
        self.__substrategy = None
        ms = self.nearest_enemy_mothership()
        if ms is not None:
            self.__substrategy = StrategyApproach(unit=self.unit, target_point=ms.coord.copy(),
                                                  distance=self.unit.gun.shot_distance)
        self._target_unit = ms
        assert self.unit is not None
        # Требуется для StrategyHarvesting
        assert hasattr(self.unit, 'elerium_stock')

    def nearest_enemy_mothership(self):
        motherships = [m for m in self.unit.scene.motherships if m.team != self.unit.team and m.is_alive]
        if motherships:
            motherships = sorted(motherships, key=lambda x: x.distance_to(self.unit))
            return motherships[0]
        return None

    @property
    def is_finished(self):
        return self.__done

    def game_step(self):
        if self.__substrategy is not None:
            self.__substrategy.game_step()
            if not self.__substrategy.is_finished:
                return
            self.__substrategy = None

        if self._target_unit and self._target_unit.is_alive:
            self.unit.gun.shot(self._target_unit)
            return

        if self._target_unit is not None and not self._target_unit.is_alive:
            ms = self.nearest_enemy_mothership()
            if ms is not None:
                self.__substrategy = StrategyApproach(unit=self.unit, target_point=ms.coord.copy(),
                                                      distance=self.unit.gun.shot_distance)
            self._target_unit = ms
            return

        if self._target_unit is None:
            self.__done = True



# -*- coding: utf-8 -*-
from unittest import TestCase

import mock

from astrobox.core import Unit
from astrobox.space_field import SpaceField
from robogame_engine.geometry import Point
from robogame_engine.theme import theme


class CargoBoxVehicle(Unit):

    def __init__(self, coord, *args, **kwargs):
        super(CargoBoxVehicle, self).__init__(*args, **kwargs)
        self.coord = coord
        self.on_load_complete = mock.MagicMock()
        self.on_unload_complete = mock.MagicMock()


class TestCargo(TestCase):

    def setUp(self):
        self.scene = SpaceField(theme_mod_path='tests.themes.for_cargo_box')
        self.initial_cargo = 50
        self.maximum_cargo = 100
        self.half_load_speed = theme.CARGO_TRANSITION_SPEED // 2
        self.unit1 = CargoBoxVehicle(
            coord=Point(0, 0), payload=self.initial_cargo, max_payload=self.maximum_cargo)
        self.unit2 = CargoBoxVehicle(
            coord=Point(0, 0), payload=self.initial_cargo, max_payload=self.maximum_cargo)

    def test_propeties(self):
        self.assertEqual(self.unit1.payload, self.initial_cargo)
        self.assertFalse(self.unit1.is_full)
        self.assertFalse(self.unit1.is_empty)
        self.assertEqual(self.unit1.free_space, self.maximum_cargo - self.initial_cargo)
        self.assertEqual(self.unit1.fullness, .5)
        self.unit1 = CargoBoxVehicle(
            coord=Point(0, 0), payload=self.maximum_cargo, max_payload=self.maximum_cargo)
        self.assertTrue(self.unit1.is_full)
        self.assertFalse(self.unit1.is_empty)
        self.assertEqual(self.unit1.free_space, 0)
        self.assertEqual(self.unit1.fullness, 1.0)
        self.unit1 = CargoBoxVehicle(
            coord=Point(0, 0), payload=0, max_payload=self.maximum_cargo)
        self.assertFalse(self.unit1.is_full)
        self.assertTrue(self.unit1.is_empty)
        self.assertEqual(self.unit1.free_space, self.maximum_cargo)
        self.assertEqual(self.unit1.fullness, 0.0)

    def test_load(self):
        self.unit1.load_from(self.unit2)
        self.unit1.game_step()
        self.assertEqual(self.unit1.payload, self.initial_cargo + theme.CARGO_TRANSITION_SPEED)
        self.assertEqual(self.unit2.payload, self.initial_cargo - theme.CARGO_TRANSITION_SPEED)
        self.assertEqual(self.unit1.on_load_complete.call_count, 0)

    def test_load_from_empty(self):
        self.unit2 = CargoBoxVehicle(
            coord=Point(0, 0), payload=0, max_payload=self.maximum_cargo)
        self.unit1.load_from(self.unit2)
        self.unit1.game_step()
        self.assertEqual(self.unit1.payload, self.initial_cargo)
        self.assertEqual(self.unit2.payload, 0)
        self.assertEqual(self.unit1.on_load_complete.call_count, 1)

    def test_load_from_nearly_empty(self):
        self.unit2 = CargoBoxVehicle(
            coord=Point(0, 0), payload=self.half_load_speed, max_payload=self.maximum_cargo)
        self.unit1.load_from(self.unit2)
        self.unit1.game_step()
        self.assertEqual(self.unit1.payload, self.initial_cargo + self.half_load_speed)
        self.assertEqual(self.unit2.payload, 0)
        self.assertEqual(self.unit1.on_load_complete.call_count, 1)

    def test_load_to_full(self):
        self.unit1 = CargoBoxVehicle(
            coord=Point(0, 0), payload=self.maximum_cargo, max_payload=self.maximum_cargo)
        self.unit1.load_from(self.unit2)
        self.unit1.game_step()
        self.assertEqual(self.unit1.payload, self.maximum_cargo)
        self.assertEqual(self.unit2.payload, self.initial_cargo)
        self.assertEqual(self.unit1.on_load_complete.call_count, 0)

    def test_load_to_almost_full(self):
        half_load_speed = theme.CARGO_TRANSITION_SPEED // 2
        self.unit1 = CargoBoxVehicle(
            coord=Point(0, 0),
            payload=self.maximum_cargo - half_load_speed,
            max_payload=self.maximum_cargo)
        self.unit1.load_from(self.unit2)
        self.unit1.game_step()
        self.assertEqual(self.unit1.payload, self.maximum_cargo)
        self.assertEqual(self.unit2.payload, self.initial_cargo - half_load_speed)

    def test_unload(self):
        self.unit1.unload_to(self.unit2)
        self.unit1.game_step()
        self.assertEqual(self.unit1.payload, self.initial_cargo - theme.CARGO_TRANSITION_SPEED)
        self.assertEqual(self.unit2.payload, self.initial_cargo + theme.CARGO_TRANSITION_SPEED)
        self.assertEqual(self.unit1.on_unload_complete.call_count, 0)

    def test_unload_from_empty(self):
        self.unit1 = CargoBoxVehicle(coord=Point(0, 0), payload=0, max_payload=self.maximum_cargo)
        self.unit1.unload_to(self.unit2)
        self.unit1.game_step()
        self.assertEqual(self.unit1.payload, 0)
        self.assertEqual(self.unit2.payload, self.initial_cargo)
        self.assertEqual(self.unit1.on_unload_complete.call_count, 0)

    def test_unload_from_nearly_empty(self):
        self.unit1 = CargoBoxVehicle(
            coord=Point(0, 0), payload=self.half_load_speed, max_payload=self.maximum_cargo)
        self.unit1.unload_to(self.unit2)
        self.unit1.game_step()
        self.assertEqual(self.unit1.payload, 0)
        self.assertEqual(self.unit2.payload, self.initial_cargo + self.half_load_speed)
        self.assertEqual(self.unit1.on_unload_complete.call_count, 1)

    def test_unload_to_full(self):
        self.unit2 = CargoBoxVehicle(
            coord=Point(0, 0), payload=self.maximum_cargo, max_payload=self.maximum_cargo)
        self.unit1.unload_to(self.unit2)
        self.unit1.game_step()
        self.assertEqual(self.unit1.payload, self.initial_cargo)
        self.assertEqual(self.unit2.payload, self.maximum_cargo)
        self.assertEqual(self.unit1.on_unload_complete.call_count, 1)

    def test_unload_to_almost_full(self):
        self.unit2 = CargoBoxVehicle(
            coord=Point(0, 0),
            payload=self.maximum_cargo - self.half_load_speed,
            max_payload=self.maximum_cargo
        )
        self.unit1.unload_to(self.unit2)
        self.unit1.game_step()
        self.assertEqual(self.unit1.payload, self.initial_cargo - self.half_load_speed)
        self.assertEqual(self.unit2.payload, self.maximum_cargo)
        self.assertEqual(self.unit1.on_unload_complete.call_count, 1)

    def test_big_distance(self):
        self.unit2 = CargoBoxVehicle(
            coord=Point(theme.CARGO_TRANSITION_DISTANCE, 0),
            payload=self.initial_cargo,
            max_payload=self.maximum_cargo
        )
        self.unit1.load_from(self.unit2)
        self.unit1.game_step()
        self.assertEqual(self.unit1.payload, self.initial_cargo)
        self.assertEqual(self.unit2.payload, self.initial_cargo)
        self.assertEqual(self.unit1.on_load_complete.call_count, 1)
        self.unit1.unload_to(self.unit2)
        self.unit1.game_step()
        self.assertEqual(self.unit1.payload, self.initial_cargo)
        self.assertEqual(self.unit2.payload, self.initial_cargo)
        self.assertEqual(self.unit1.on_unload_complete.call_count, 1)



# -*- coding: utf-8 -*-
from robogame_engine.theme import theme


class Cargo(object):
    __payload = 0
    __max_volume = 0

    def __init__(self, owner, payload=0, max_payload=1):
        self.__owner = owner
        if max_payload < 1:
            raise ValueError("max_payload should be greater than 0")
        self.__payload = min(payload, max_payload)
        self.__max_payload = max_payload

    def __str__(self):
        return '[{}]{} payload {}/{}'.format(id(self), self.__class__.__name__, self.__payload, self.__max_payload)

    def _clip_payload(self, batch):
        if self.__payload < batch:
            batch = self.__payload
        self.__payload -= batch
        return batch

    def _transfer_payload(self, batch, cargo_from):
        batched = cargo_from._clip_payload(batch)
        self.__payload += batched
        return batched

    @property
    def owner(self):
        return self.__owner

    @property
    def payload(self):
        return self.__payload

    @property
    def fullness(self):
        return self.__payload / float(self.__max_payload)

    @property
    def is_empty(self):
        return self.__payload <= 0

    @property
    def is_full(self):
        return self.__payload >= self.__max_payload

    @property
    def free_space(self):
        return self.__max_payload - self.__payload


class CargoTransition(object):
    selectable = False

    def __init__(self, cargo_from=None, cargo_to=None):
        super(CargoTransition, self).__init__()
        self.cargo_from = cargo_from
        self.cargo_to = cargo_to
        self.__distance = theme.CARGO_TRANSITION_DISTANCE
        self.__batch_processed = 0
        self.__transition_limit = max(min(self.cargo_to.free_space, self.cargo_from.payload), 0)
        self.__transition_speed = theme.CARGO_TRANSITION_SPEED
        if self.__transition_speed < 1:
            raise Exception("transition_speed should be greater than 0")
        self.__done = False
        self.__was_transfer = False

    @property
    def is_finished(self):
        return self.__done

    @property
    def was_transfer(self):
        return self.__was_transfer

    def game_step(self):
        self.__was_transfer = False
        # Ограничиваем дистанцию переноса
        if self.cargo_to.owner.distance_to(self.cargo_from.owner) >= self.__distance:
            self.__done = True
            return
        # Берем максимально возможный кусок который можем переместить за такт
        batch = min(self.__transition_speed,
                    self.__transition_limit - self.__batch_processed,
                    self.cargo_to.free_space, self.cargo_from.payload)
        if batch <= 0:
            self.__done = True
            return
        self.__batch_processed += self.cargo_to._transfer_payload(batch, self.cargo_from)
        self.__was_transfer = True
        if self.cargo_from.is_empty or self.cargo_to.is_full:
            self.__done = True
